package qtree

import (
	"math"

	"github.com/BTrDB/btrdb-server/internal/bstore"
)

// sample variance
func (n *QTreeNode) OpCountVariance(mean float64) float64 {
	total := 0.0
	cnt := uint64(0)
	if n.isLeaf {
		for i := 0; i < int(n.vector_block.Len); i++ {
			total += (n.vector_block.Value[i] - mean) * (n.vector_block.Value[i] - mean)
		}
		if n.vector_block.Len <= 1 {
			return 0
		}
		return total / float64(n.vector_block.Len-1)
	} else {
		//general online algorithm
		flag := false
		accMean := 0.0
		accVariance := 0.0
		for i := 0; i < bstore.KFACTOR; i++ {
			if n.core_block.Count[i] == 0 {
				continue
			}
			if !flag {
				accMean = n.core_block.Mean[i]
				accVariance = n.core_block.Variance[i]
				flag = true
				cnt += n.core_block.Count[i]
				continue
			}
			// cnt is an accumulator on the lefthand side, coreblock is righthand side
			delta := accMean - n.core_block.Mean[i]
			m_a := accVariance * float64(cnt-1)
			m_b := n.core_block.Variance[i] * float64(n.core_block.Count[i]-1)
			totalCount := cnt + n.core_block.Count[i]
			m2 := m_a + m_b + (delta*delta)*(float64(cnt*n.core_block.Count[i])/float64(totalCount))
			accVariance = m2 / float64(totalCount-1)
			accMean = ((accMean * float64(cnt)) + (n.core_block.Mean[i] * float64(n.core_block.Count[i]))) / float64(totalCount)
			cnt = totalCount
		}
		return accVariance
	}
}

func (n *QTreeNode) OpCountMean() (uint64, float64) {
	total := 0.0
	cnt := uint64(0)
	if n.isLeaf {
		for i := 0; i < int(n.vector_block.Len); i++ {
			total += n.vector_block.Value[i]
		}
		return uint64(n.vector_block.Len), total / float64(n.vector_block.Len)
	} else {
		for i := 0; i < bstore.KFACTOR; i++ {
			if n.core_block.Count[i] == 0 {
				continue
			}
			cnt += n.core_block.Count[i]
			total += n.core_block.Mean[i] * float64(n.core_block.Count[i])
		}
		return cnt, total / float64(cnt)
	}
}

func (n *QTreeNode) OpMin() float64 {
	if n.isLeaf {
		min := n.vector_block.Value[0]
		for i := 0; i < int(n.vector_block.Len); i++ {
			if n.vector_block.Value[i] < min {
				min = n.vector_block.Value[i]
			}
		}
		return min
	} else {
		min := float64(0)
		minset := false
		for i := 0; i < len(n.core_block.Min); i++ {
			if n.core_block.Count[i] == 0 {
				continue
			}
			if !minset || n.core_block.Min[i] < min {
				min = n.core_block.Min[i]
				minset = true
			}
		}
		return min
	}
}

func (n *QTreeNode) OpMax() float64 {
	if n.isLeaf {
		max := n.vector_block.Value[0]
		for i := 0; i < int(n.vector_block.Len); i++ {
			if n.vector_block.Value[i] > max {
				max = n.vector_block.Value[i]
			}
		}
		return max
	} else {
		max := float64(0)
		maxset := false
		for i := 0; i < len(n.core_block.Max); i++ {
			if n.core_block.Count[i] == 0 {
				continue
			}
			if !maxset || n.core_block.Max[i] > max {
				max = n.core_block.Max[i]
				maxset = true
			}
		}
		return max
	}
}

/*

ok so here is the problem. If we call opreduce on a core node, then we can only deliver
pointwidths GREATER than our pointwidth and less than pointwidth + 6 right?
but as a leaf we can potentially deliver pointwidths down to 0...
*/
func (n *QTreeNode) OpReduce(pointwidth uint8, index uint64) (uint64, float64, float64, float64, float64) {
	if !n.isLeaf && pointwidth < n.PointWidth() {
		lg.Panic("Bad pointwidth for core. See code comment")
	}
	if pointwidth > n.PointWidth()+PWFACTOR {
		lg.Panic("Can't guarantee this PW")
	}
	maxpw := n.PointWidth() + PWFACTOR
	pwdelta := pointwidth - n.PointWidth()
	width := int64(1) << pointwidth
	maxidx := 1 << (maxpw - pointwidth)
	if maxidx <= 0 || index >= uint64(maxidx) {
		lg.Critical("node is %s", n.TreePath())
		lg.Panic("bad index", maxidx, index)
	}
	sum := 0.0
	min := math.NaN()
	max := math.NaN()
	minset := false
	maxset := false
	varianceSum := 0.0
	mean := 0.0
	count := uint64(0)
	if n.isLeaf {
		st := n.StartTime() + int64(index)*width
		et := st + width
		if n.vector_block.Len != 0 {
			for i := 0; i < int(n.vector_block.Len); i++ {
				if n.vector_block.Time[i] < st {
					continue
				}
				if n.vector_block.Time[i] >= et {
					break
				}
				v := n.vector_block.Value[i]
				sum += v
				if !minset || v < min {
					minset = true
					min = v
				}
				if !maxset || v > max {
					maxset = true
					max = v
				}
				count++
			}
			mean = sum / float64(count)
			for i := 0; i < int(n.vector_block.Len); i++ {
				if n.vector_block.Time[i] < st {
					continue
				}
				if n.vector_block.Time[i] >= et {
					break
				}
				v := n.vector_block.Value[i]
				temp := v - mean
				varianceSum += temp * temp
			}
		}
		if count <= 1 {
			return count, min, mean, max, 0
		}

		return count, min, mean, max, varianceSum / float64(count-1)
	} else {
		s := index << pwdelta
		e := (index + 1) << pwdelta
		flag := false
		for i := s; i < e; i++ {
			if n.core_block.Count[i] == 0 {
				continue
			}
			if !flag {
				varianceSum = n.core_block.Variance[i]
				count += n.core_block.Count[i]
				mean = n.core_block.Mean[i]
				sum += n.core_block.Mean[i] * float64(n.core_block.Count[i])
				if !minset || n.core_block.Min[i] < min {
					minset = true
					min = n.core_block.Min[i]
				}
				if !maxset || n.core_block.Max[i] > max {
					maxset = true
					max = n.core_block.Max[i]
				}
				flag = true
				continue
			}

			//cnt is a lefthand side accumulator, coreblock is the right hand side
			delta := mean - n.core_block.Mean[i]
			m_a := varianceSum * float64(count-1)
			m_b := n.core_block.Variance[i] * float64(n.core_block.Count[i]-1)
			totalCount := count + n.core_block.Count[i]
			m2 := m_a + m_b + (delta*delta)*(float64(count*n.core_block.Count[i])/float64(totalCount))
			varianceSum = m2 / float64(totalCount-1)
			mean = ((mean * float64(count)) + (n.core_block.Mean[i] * float64(n.core_block.Count[i]))) / float64(totalCount)
			count = totalCount //+= n.core_block.Count[i]
			sum += n.core_block.Mean[i] * float64(n.core_block.Count[i])
			if !minset || n.core_block.Min[i] < min {
				minset = true
				min = n.core_block.Min[i]
			}
			if !maxset || n.core_block.Max[i] > max {
				maxset = true
				max = n.core_block.Max[i]
			}
		}
		return count, min, mean, max, varianceSum
	}
}
